@@ -39,9 +39,13 @@ typedef struct yaspot_view_t {
     GtkWidget *image;
 
     /* for play queue timer */
+    GtkTreeSelection *selection;
+    gint current_playback_row;
     GtkProgressBar *timer;
-    GThread *timer_thread;
+    GThread *ui_thread;
     GAsyncQueue *timer_queue;
+    HildonButton *button_pp;
+    gboolean old_paused;
 
     GList *artist_view;
     progress_t progress;
@@ -54,7 +58,11 @@ static gint artist_view_block = FALSE;
 
 /* *** common *** */
 
-static const guint32 STOP_TIMER_LOOP = 1;
+static const gint32 STOP_TIMER_LOOP = -1;
+static const gint32 UPDATE_SELECTION = -2;
+static const gint32 UPDATE_PLAYPAUSE = -3;
+
+static void state_changed_cb(gboolean playing, gboolean paused, gpointer data);
 
 static yaspot_view_t* view_new(userdata_t *u) {
     yaspot_view_t *v;
@@ -98,13 +106,15 @@ static void queue_destroy(GtkWidget *widget, GdkEvent *event, gpointer data) {
 
     if (v->timer) {
         yaspot_glue_set_timer_cb(v->u->session, NULL, NULL);
-        g_async_queue_push(v->timer_queue, GUINT_TO_POINTER(STOP_TIMER_LOOP));
+        g_async_queue_push(v->timer_queue, GINT_TO_POINTER(STOP_TIMER_LOOP));
         g_print("queue_destroy: wait for thread to terminate\n");
-        g_thread_join(v->timer_thread);
+        g_thread_join(v->ui_thread);
 
         g_async_queue_unref(v->timer_queue);
     }
 
+    yaspot_glue_del_state_cb(v->u->session, state_changed_cb, v);
+
     destroy(widget, event, data);
 }
 
@@ -683,31 +693,6 @@ void yaspot_show_album(userdata_t *u, yaspot_glue_response_t *list, gint num) {
     gtk_widget_show_all(v->window);
 }
 
-
-static void queue_row_activated(GtkTreeView *treeview, GtkTreePath *path, GtkTreeViewColumn *column, gpointer data) {
-    yaspot_view_t *v = (yaspot_view_t*)data;
-    yaspot_glue_response_t *r;
-    GtkTreeIter iter;
-    GtkTreeModel *model;
-    gint value;
-
-    model = gtk_tree_view_get_model(treeview);
-    if (!gtk_tree_model_get_iter(model, &iter, path)) {
-        g_print("queue: failed to get iterator\n");
-        return;
-    }
-
-    gtk_tree_model_get(model, &iter, PLAYLIST_COL_TRACK, &value, -1);
-
-    if (value > -1) {
-        if ((r = yaspot_glue_current_queue(v->u->session))) {
-            g_print("play song %d\n", value);
-            yaspot_glue_play(r, value);
-            yaspot_glue_response_unref(r);
-        }
-    }
-}
-
 static void update_timer(yaspot_glue_t *t, gint now, gint length, gpointer userdata) {
     yaspot_view_t *v = (yaspot_view_t*)userdata;
     guint32 msg;
@@ -719,12 +704,44 @@ static void update_timer(yaspot_glue_t *t, gint now, gint length, gpointer userd
 
     msg = (guint32)now + ((guint32)length << 16);
 
-    g_async_queue_push(v->timer_queue, GUINT_TO_POINTER((guint32)msg));
+    g_async_queue_push(v->timer_queue, GINT_TO_POINTER((guint32)msg));
+}
+
+static void update_selection(yaspot_view_t *v, gboolean scroll) {
+    GtkTreeModel *model;
+    GtkTreeIter iter;
+    gint new_row;
+    gint scroll_pos;
+
+    g_assert(v);
+    g_assert(v->selection);
+    g_assert(v->treeview);
+
+    new_row = yaspot_glue_current_track_num(v->u->session);
+
+    g_print("update_selection: old row %d new row %d\n", v->current_playback_row, new_row);
+
+    if (v->current_playback_row == new_row)
+        return;
+
+    v->current_playback_row = new_row;
+
+    model = gtk_tree_view_get_model(v->treeview);
+    gtk_tree_model_iter_nth_child(model, &iter, NULL, v->current_playback_row - 1);
+
+    gtk_tree_selection_select_iter(v->selection, &iter);
+
+    /* FIXME MAGIC NUMBER */
+    scroll_pos = v->current_playback_row * 70 + 35;
+    if (scroll)
+        hildon_pannable_area_scroll_to(v->area, -1, scroll_pos);
+    else
+        hildon_pannable_area_jump_to(v->area, -1, scroll_pos);
 }
 
 /* use two unsigned 16bit values in 32bit variable, pass as pointer.
  * value 1 stops thread */
-static void* timer_thread_loop(gpointer userdata) {
+static void* ui_thread_loop(gpointer userdata) {
     yaspot_view_t *v = (yaspot_view_t*)userdata;
     guint32 msg;
     gint now;
@@ -732,19 +749,35 @@ static void* timer_thread_loop(gpointer userdata) {
     gdouble fraction;
     gint loop = 1;
     gchar text[32];
+    gboolean paused;
     g_assert(v);
 
     g_async_queue_ref(v->timer_queue);
 
     while (loop) {
-        msg = (guint32)GPOINTER_TO_UINT(g_async_queue_pop(v->timer_queue));
+        msg = (gint32)GPOINTER_TO_INT(g_async_queue_pop(v->timer_queue));
 
         if (msg == STOP_TIMER_LOOP) {
             loop = 0;
+        } else if (msg == UPDATE_SELECTION) {
+
+            update_selection(v, TRUE);
+
+        } else if (msg == UPDATE_PLAYPAUSE) {
+
+            paused = !yaspot_glue_playing(v->u->session);
+            if (paused != v->old_paused) {
+                if (paused)
+                    hildon_button_set_image(v->button_pp, gtk_image_new_from_stock(GTK_STOCK_MEDIA_PLAY, GTK_ICON_SIZE_BUTTON));
+                else
+                    hildon_button_set_image(v->button_pp, gtk_image_new_from_stock(GTK_STOCK_MEDIA_PAUSE, GTK_ICON_SIZE_BUTTON));
+                v->old_paused = paused;
+            }
+
         } else {
 
-            now = (guint16)msg;
-            length = (guint32)msg >> 16;
+            now = (gint16)msg;
+            length = (gint32)msg >> 16;
 
             if (length/60 > 10)
                 g_snprintf(text, 32, "%02d:%02d [%02d:%02d]", now/60, now%60, length/60, length%60);
@@ -760,20 +793,69 @@ static void* timer_thread_loop(gpointer userdata) {
 
     g_async_queue_unref(v->timer_queue);
 
-    g_print("timer_thread_loop: done.\n");
+    g_print("ui_thread_loop: done.\n");
 
     return NULL;
 }
 
+static void queue_selection_cb(GtkTreeSelection *selection, gpointer data) {
+    yaspot_view_t *v = (yaspot_view_t*)data;
+    yaspot_glue_response_t *r;
+    GtkTreeIter iter;
+    GtkTreeModel *model;
+    gint value;
+
+    g_assert(selection);
+    g_assert(v);
+
+    model = gtk_tree_view_get_model(v->treeview);
+    if (!gtk_tree_selection_get_selected(selection, NULL, &iter)) {
+        g_print("queue_selection_cb: failed to get iter\n");
+    }
+
+    gtk_tree_model_get(model, &iter, PLAYLIST_COL_TRACK, &value, -1);
+
+    if (value == v->current_playback_row) {
+        return;
+    }
+
+    if (value > -1) {
+        if ((r = yaspot_glue_current_queue(v->u->session))) {
+            g_print("play song %d\n", value);
+            yaspot_glue_play(r, value);
+            yaspot_glue_response_unref(r);
+            update_selection(v, TRUE);
+        }
+    }
+}
+
+static void state_changed_cb(gboolean playing, gboolean paused, gpointer data) {
+    yaspot_view_t *v = (yaspot_view_t*)data;
+
+    g_assert(v);
+
+    g_print("queue: state changed\n");
+
+    if (!playing)
+        return;
+
+    g_async_queue_push(v->timer_queue, GINT_TO_POINTER(UPDATE_PLAYPAUSE));
+
+    if (!paused)
+        g_async_queue_push(v->timer_queue, GINT_TO_POINTER(UPDATE_SELECTION));
+}
+
 void yaspot_show_queue(userdata_t *u) {
     yaspot_view_t *v;
     yaspot_glue_response_t *r;
     GtkListStore *model;
     GtkWidget *vbox;
     GtkWidget *status_hbox;
-    GtkWidget *button_prev;
-    GtkWidget *button_pp;
-    GtkWidget *button_next;
+    HildonButton *button_prev;
+    HildonButton *button_next;
+    GtkWidget *image_prev;
+    GtkWidget *image_next;
+    GtkWidget *image_pp;
 
     v = view_new(u);
     v->window = hildon_stackable_window_new();
@@ -785,32 +867,41 @@ void yaspot_show_queue(userdata_t *u) {
     status_hbox = gtk_hbox_new(FALSE, 10);
     v->timer = GTK_PROGRESS_BAR( gtk_progress_bar_new() );
 
-    button_prev = hildon_button_new_with_text(HILDON_SIZE_AUTO_WIDTH | HILDON_SIZE_FINGER_HEIGHT,
-                                              HILDON_BUTTON_ARRANGEMENT_VERTICAL,
-                                              "<<",
-                                              NULL);
-    button_pp   = hildon_button_new_with_text(HILDON_SIZE_AUTO_WIDTH | HILDON_SIZE_FINGER_HEIGHT,
-                                              HILDON_BUTTON_ARRANGEMENT_VERTICAL,
-                                              "|>",
-                                              NULL);
-    button_next = hildon_button_new_with_text(HILDON_SIZE_AUTO_WIDTH | HILDON_SIZE_FINGER_HEIGHT,
-                                              HILDON_BUTTON_ARRANGEMENT_VERTICAL,
-                                              ">>",
-                                              NULL);
-    gtk_widget_set_size_request(button_prev, 100, -1);
-    gtk_widget_set_size_request(button_pp, 100, -1);
-    gtk_widget_set_size_request(button_next, 100, -1);
+    button_prev = hildon_button_new(HILDON_SIZE_AUTO_WIDTH | HILDON_SIZE_FINGER_HEIGHT,
+                                    HILDON_BUTTON_ARRANGEMENT_VERTICAL);
+    v->button_pp= hildon_button_new(HILDON_SIZE_AUTO_WIDTH | HILDON_SIZE_FINGER_HEIGHT,
+                                    HILDON_BUTTON_ARRANGEMENT_VERTICAL);
+    button_next = hildon_button_new(HILDON_SIZE_AUTO_WIDTH | HILDON_SIZE_FINGER_HEIGHT,
+                                    HILDON_BUTTON_ARRANGEMENT_VERTICAL);
+    gtk_widget_set_size_request(GTK_WIDGET(button_prev), 100, -1);
+    gtk_widget_set_size_request(GTK_WIDGET(v->button_pp), 100, -1);
+    gtk_widget_set_size_request(GTK_WIDGET(button_next), 100, -1);
 
-    gtk_box_pack_start(GTK_BOX(status_hbox), button_prev, FALSE, FALSE, 0);
-    gtk_box_pack_start(GTK_BOX(status_hbox), button_pp, FALSE, FALSE, 0);
-    gtk_box_pack_start(GTK_BOX(status_hbox), button_next, FALSE, FALSE, 0);
-    gtk_box_pack_start(GTK_BOX(status_hbox), v->timer, TRUE, TRUE, 0);
+    gtk_box_pack_start(GTK_BOX(status_hbox), GTK_WIDGET(button_prev), FALSE, FALSE, 0);
+    gtk_box_pack_start(GTK_BOX(status_hbox), GTK_WIDGET(v->button_pp), FALSE, FALSE, 0);
+    gtk_box_pack_start(GTK_BOX(status_hbox), GTK_WIDGET(button_next), FALSE, FALSE, 0);
+    gtk_box_pack_start(GTK_BOX(status_hbox), GTK_WIDGET(v->timer), TRUE, TRUE, 0);
 
-    gtk_box_pack_start(GTK_BOX(vbox), v->area, TRUE, TRUE, 0);
+    gtk_box_pack_start(GTK_BOX(vbox), GTK_WIDGET(v->area), TRUE, TRUE, 0);
     gtk_box_pack_start(GTK_BOX(vbox), status_hbox, FALSE, FALSE, 0);
 
     gtk_container_add(GTK_CONTAINER(v->window), GTK_WIDGET(vbox));
 
+    image_prev = gtk_image_new_from_stock(GTK_STOCK_MEDIA_PREVIOUS, GTK_ICON_SIZE_BUTTON);
+    image_next = gtk_image_new_from_stock(GTK_STOCK_MEDIA_NEXT, GTK_ICON_SIZE_BUTTON);
+    if (yaspot_glue_playing(u->session))
+        image_pp = gtk_image_new_from_stock(GTK_STOCK_MEDIA_PAUSE, GTK_ICON_SIZE_BUTTON);
+    else
+        image_pp = gtk_image_new_from_stock(GTK_STOCK_MEDIA_PLAY, GTK_ICON_SIZE_BUTTON);
+
+    v->old_paused = !yaspot_glue_playing(u->session);
+
+    hildon_button_set_image(button_prev, image_prev);
+    hildon_button_set_image(button_next, image_next);
+    hildon_button_set_image(v->button_pp, image_pp);
+
+    /* FIXME are these images freed with window etc? */
+
     v->u = u;
 
     gtk_window_set_title(GTK_WINDOW(v->window), _("Playing Queue"));
@@ -820,26 +911,34 @@ void yaspot_show_queue(userdata_t *u) {
     if ((r = yaspot_glue_current_queue(u->session))) {
         yaspot_glue_response_to_queue_model(r, model);
         v->r = r;
+
+        v->selection = gtk_tree_view_get_selection(v->treeview);
+        gtk_tree_selection_set_mode(v->selection, GTK_SELECTION_BROWSE);
+
     } else {
         gtk_list_store_clear(model);
     }
 
     g_signal_connect(button_prev, "clicked", G_CALLBACK(msp_prev_cb), u);
-    g_signal_connect(button_pp, "clicked", G_CALLBACK(msp_playpause_cb), u);
+    g_signal_connect(v->button_pp, "clicked", G_CALLBACK(msp_playpause_cb), u);
     g_signal_connect(button_next, "clicked", G_CALLBACK(msp_next_cb), u);
 
     g_signal_connect(G_OBJECT(v->window), "delete_event",
                      G_CALLBACK(queue_destroy), v);
 
-    g_signal_connect(G_OBJECT(v->treeview), "row-activated", G_CALLBACK(queue_row_activated), v);
-
     gtk_widget_show_all(v->window);
     gtk_widget_show_all(GTK_WIDGET(v->timer));
     if (!v->timer_queue)
         v->timer_queue = g_async_queue_new();
-    if (!v->timer_thread)
-        v->timer_thread = g_thread_create(timer_thread_loop, v, TRUE, NULL);
+    if (!v->ui_thread)
+        v->ui_thread = g_thread_create(ui_thread_loop, v, TRUE, NULL);
     yaspot_glue_set_timer_cb(u->session, update_timer, v);
+
+    if (v->r) {
+        update_selection(v, FALSE);
+        g_signal_connect(G_OBJECT(v->selection), "changed", G_CALLBACK(queue_selection_cb), v);
+        yaspot_glue_add_state_cb(u->session, state_changed_cb, v);
+    }
 }
 
 void yaspot_show_biography(userdata_t *u, yaspot_glue_response_t *r) {
